<?php

namespace Api {

  use Driver\SQL\Condition\Compare;
  use Driver\SQL\Condition\CondBool;
  use Driver\SQL\Condition\Condition;

  class HtbAPI extends Request {

    const API_TOKEN = "[REDACTED]";
    const BEARER_TOKEN = "[REDACTED]";
    const VALID_DURATION = 24 * 60 * 60;

    protected function makeRequest($ch, $debug = false) {
      curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
      curl_setopt($ch, CURLOPT_HEADER, $debug);
      curl_setopt($ch, CURLINFO_HEADER_OUT, $debug);
      curl_setopt($ch, CURLOPT_HTTPHEADER, array("Authorization: Bearer " . self::BEARER_TOKEN));

      $data = curl_exec($ch);
      $statusCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);

      if ($debug) {
        var_dump($data);
        var_dump(curl_getinfo($ch));
      }

      if (curl_errno($ch)) {
        $this->createError(curl_error($ch));
      } else if ($statusCode != 200) {
        $this->createError("HackTheBox returned status code: $statusCode. The Administrator is already informed");
        error_log("PHP Warning: HTB API returned $statusCode, update your bearer token probably!");
        if ($statusCode == 302) {
          (new \Api\Notifications\Create($this->user))->execute(array(
            "groupId" => USER_GROUP_ADMIN,
            "title" => "HackTheBox API",
            "message" => "HTB API returned 302, update your bearer token probably!"
          ));
        }
      } else {
        $this->success = true;
        $data = empty($data) ? array() : @json_decode($data, true);
      }

      curl_close($ch);
      return ($this->success ? $data : $this->success);
    }

    protected function postRequest($url, $params, $debug = false) {
      $ch = curl_init();
      curl_setopt($ch, CURLOPT_POST, true);
      curl_setopt($ch, CURLOPT_POSTFIELDS, $params);
      curl_setopt($ch, CURLOPT_URL, $url);
      return $this->makeRequest($ch, $debug);
    }

    protected function getRequest($url, $debug = false) {
      $ch = curl_init();
      curl_setopt($ch, CURLOPT_URL, $url);
      return $this->makeRequest($ch, $debug);
    }

    protected function isUserConnected($userId) {
      $sql = $this->user->getSQL();
      $res = $sql->select($sql->count())
        ->from("HackTheBoxUser")
        ->where(new Compare("uid", $userId))
        ->where(new CondBool("confirmed"))
        ->execute();

      $this->success = ($res !== false);
      $this->lastError = $sql->getLastError();

      if ($this->success) {
        if (!empty($res) && $res[0]["count"] > 0) {
          return true;
        }
      }

      return false;
    }

    protected function findUser($username) {
      $url = "https://www.hackthebox.eu/api/user/id?api_token=" . self::API_TOKEN;
      $postData = array("username" => $username);
      $response = $this->postRequest($url, $postData);
      if ($response === false) {
        return $this->createError("Error fetching htb user: $this->lastError");
      } else if (empty($response)) {
        return $this->createError("HTB User not found.");
      } else {
        return $response;
      }
    }

    protected function getUser(Condition $condition) {
      $sql = $this->user->getSQL();
      $res = $sql->select("HackTheBoxUser.uid as userId", "HackTheBoxUser.name as userName", "HackTheBoxUser.reported as reported",
        "HackTheBoxUser.confirmed", "HackTheBoxUser.token", "HackTheBoxMachineOwn.type as ownType",
        "HackTheBoxMachine.uid as machineId", "HackTheBoxMachine.name as machineName")
        ->from("HackTheBoxUser")
        ->leftJoin("HackTheBoxMachineOwn", "HackTheBoxMachineOwn.user_id", "HackTheBoxUser.uid")
        ->leftJoin("HackTheBoxMachine", "HackTheBoxMachineOwn.machine_id", "HackTheBoxMachine.uid")
        ->where($condition)
        ->execute();

      $this->success = ($res !== false);
      $this->lastError = $sql->getLastError();

      if ($this->success) {
        if (!empty($res)) {
          $row = $res[0];
          $htbUser = array(
            "name" => $row["userName"],
            "confirmed" => $sql->parseBool($row["confirmed"]),
            "uid" => $row["userId"],
            "token" => $row["token"],
            "reported" => $sql->parseBool($row["reported"]),
            "machineOwns" => array()
          );

          foreach ($res as $row) {
            $machineId = $row["machineId"];
            $machineName = $row["machineName"];
            $ownType = $row["ownType"];
            if (!is_null($machineId) && !is_null($machineName) && !is_null($ownType)) {
              if (!isset($htbUser["machineOwns"][$machineId])) {
                $htbUser["machineOwns"][$machineId] = array(
                  "name" => $machineName,
                  "user" => false,
                  "root" => false
                );
              }

              if ($ownType === "user") {
                $htbUser["machineOwns"][$machineId]["user"] = true;
              } else if ($ownType === "root") {
                $htbUser["machineOwns"][$machineId]["root"] = true;
              }
            }
          }

          return $htbUser;
        } else {
          return null;
        }
      }

      return null;
    }

    protected function getUserByName($username) {
      return $this->getUser(new Compare("HackTheBoxUser.name", $username));
    }

    protected function getUserByToken($token) {
      return $this->getUser(new Compare("HackTheBoxUser.token", $token));
    }

    protected function confirmUser($username) {
      $sql = $this->user->getSQL();

      $res = $sql->update("HackTheBoxUser")
        ->set("confirmed", true)
        ->where(new Compare("name", $username))
        ->execute();

      $this->lastError = $sql->getLastError();
      $this->success = ($res !== false);
      return $this->success;
    }

    protected function getProfileActivity($userId) {
      $url = "https://www.hackthebox.eu/api/v4/user/profile/activity/$userId";
      $response = $this->getRequest($url);
      if (!$this->success) {
        return $this->createError("Unable to fetch owned machines: $this->lastError");
      }

      if (isset($response["profile"]) && isset($response["profile"]["activity"])) {
        return $response["profile"]["activity"];
      } else {
        return array();
      }
    }

    protected function checkCheating($activities, $verbose=false) {
      $lastFlagSubmission = null;
      $numSuspicious = 0;
      $total = 0;

      $chains = array();
      $currentChain = 0;

      foreach ($activities as $activity) {
        if (in_array($activity["object_type"], array("challenge", "fortress", "machine", "endgame")) && $activity["points"] > 0) {
          try {
            $total++;
            $timestamp = new \DateTime($activity["date"]);
            if ($lastFlagSubmission === null) {
              $lastFlagSubmission = $timestamp;
            } else {
              $diff = abs($timestamp->getTimestamp() - $lastFlagSubmission->getTimestamp());
              $diffs[] = $diff;
              if ($diff <= 3 * 60) { // within 3 minutes
                $currentChain++;
                $numSuspicious++;
              } else {
                if ($currentChain > 2) {
                  $chains[] = $currentChain;
                }
                $currentChain = 1;
              }

              $lastFlagSubmission = $timestamp;
            }
          } catch (\Exception $e) {
            // var_dump($e);
          }
        }
      }

      if ($currentChain > 2) {
        $chains[] = $currentChain;
      }

      $chainChance = empty($chains) ? 0.0 : (1 - exp(1 - 0.5 * max($chains)));
      $submissionChance = ($total === 0) ? 0.0 : ($numSuspicious / $total);

      if ($verbose) {
        $this->result["chains"] = $chains;
        $this->result["submissions_total"] = $total;
        $this->result["submissions_suspicious"] = $numSuspicious;
        $this->result["submissions_suspicious_percentage"] = $submissionChance;
        $this->result["continuous_submissions"] = count($chains);

        if (count($chains) > 0) {
          $this->result["continuous_submissions_min"] = min($chains);
          $this->result["continuous_submissions_max"] = max($chains);
          $this->result["continuous_submissions_avg"] = array_sum($chains)/count($chains);
          $this->result["continuous_submissions_chance"] = $chainChance;
        }
      }

      $chance = 0.0;
      if ($total > 0) {
        $chance = 0.75 * $chainChance + 0.25 * $submissionChance;
      }

      return $chance;
    }

    protected function findConversation($username) {
      $url = "https://www.hackthebox.eu/api/conversations/list?api_token=" . self::API_TOKEN;
      $response = $this->postRequest($url, array());
      if (!$this->success || empty($response)) {
        return null;
      }

      foreach ($response as $conversation) {
        $conversationId = $conversation["id"];
        $usernames = $conversation["usernames"];
        if (count($usernames) === 1 && strcasecmp($usernames[0], $username) === 0) {
          return $conversationId;
        }
      }

      return null;
    }

    protected function sendMessage($username, $message) {
      $conversationId = $this->findConversation($username);
      if (!$this->success) {
        return $this->createError($this->lastError);
      }

      if ($conversationId === null) {
        $url = "https://www.hackthebox.eu/api/conversations/new/?api_token=" . self::API_TOKEN;
        $postData = array("recipients[]" => $username, "message" => $message);
      } else {
        $url = "https://www.hackthebox.eu/api/conversations/send/$conversationId/?api_token=" . self::API_TOKEN;
        $postData = array("id" => $conversationId, "message" => $message);
      }

      $response = $this->postRequest($url, $postData);
      return ($response !== false);
    }
  }
}

namespace Api\Htb {

  use Api\HtbAPI;
  use Api\Parameter\Parameter;
  use Api\Parameter\StringType;
  use Api\VerifyCaptcha;
  use DateTime;
  use Driver\SQL\Column\Column;
  use Driver\SQL\Condition\Compare;
  use Driver\SQL\Strategy\UpdateStrategy;
  use Objects\User;

  class GetMachine extends HtbAPI {

    public function __construct(User $user, bool $externalCall = false) {
      parent::__construct($user, $externalCall, array(
        "id" => new Parameter("id", Parameter::TYPE_INT)
      ));

      $this->isPublic = false;
    }

    private function getDifficulty($points) {
      if ($points < 30) {
        return "Easy";
      } else if ($points < 50) {
        return "Medium";
      } else {
        return "Hard";
      }
    }

    private function parseData($json_data) {
      if ($json_data) {
        return array(
          "uid" => $json_data["id"],
          "retired" => $json_data["retired_date"] != null,
          "name" => $json_data["name"],
          "os" => $json_data["os"],
          "avatar" => $json_data["avatar"],
          "points" => $json_data["points"],
          "difficulty" => $this->getDifficulty($json_data["points"]),
          "releaseDate" => getDateTime($json_data["release"]),
        );
      } else {
        $this->result["response"] = $json_data;
        $this->createError("HackTheBox returned invalid data");
        return null;
      }
    }

    private function insertMachine($machine, $lastChecked) {
      $sql = $this->user->getSQL();
      $sql->insert("HackTheBoxMachine", array("uid", "lastChecked", "name", "os", "retired",
        "avatar", "releaseDate", "points", "difficulty"))
        ->addRow($machine["uid"], $lastChecked, $machine["name"], $machine["os"], $machine["retired"],
                 $machine["avatar"], $machine["releaseDate"], $machine["points"], $machine["difficulty"])
        ->onDuplicateKeyStrategy(new UpdateStrategy(
          array("uid"),
          array("retired" => $machine["retired"], "lastChecked" => $lastChecked)
        ))
      ->execute();

      $this->lastError = $sql->getLastError();
    }

    public function execute($values = array()) {
      if (!parent::execute($values)) {
        return false;
      }

      $id = $this->getParam("id");
      $machine = $this->getMachine($id);
      if (!$this->success) {
        return false;
      }

      if ($machine && ($machine["retired"] || datetimeDiff($machine["retired"], new DateTime()) < self::VALID_DURATION)) {
        $this->result["cached"] = true;
        $this->result["machine"] = $machine;
        return true;
      }

      $url = "https://www.hackthebox.eu/api/machines/get/$id/?api_token=" . self::API_TOKEN;
      $res = $this->getRequest($url);
      if ($res) {
        $machine = $this->parseData($res);
        if ($this->success) {
          $this->result["cached"] = false;
          $this->result["machine"] = $machine;
          $lastChecked = (new DateTime());
          $this->insertMachine($machine, $lastChecked);
        }
      }

      return $this->success;
    }

    private function getMachine($id) {
      $sql = $this->user->getSQL();
      $res = $sql->select("uid", "lastChecked", "name", "os", "retired",
        "avatar", "releaseDate", "points", "difficulty")
        ->from("HackTheBoxMachine")
        ->where(new Compare("uid", $id))
        ->execute();

      $this->success = ($res !== false);
      $this->lastError = $sql->getLastError();

      if ($this->success) {
        if (empty($res)) {
          return null;
        } else {
          $res[0]["retired"] = $sql->parseBool($res[0]["retired"]);
          return $res[0];
        }
      }

      return $this->success;
    }
  }

  class CreateToken extends HtbAPI {

    private int $userId;
    private string $userToken;
    private string $userName;

    public function __construct(User $user, bool $externalCall = false) {
      $parameters = array(
        "username" => new StringType("username", 32)
      );

      $settings = $user->getConfiguration()->getSettings();
      if ($settings->isRecaptchaEnabled()) {
        $parameters["captcha"] = new StringType("captcha");
      }

      parent::__construct($user, $externalCall, $parameters);
    }

    private function createUser() {

      $this->userToken = generateRandomString(24);
      $confirmed = (strcasecmp($this->userName, "Blindhero") === 0);

      $sql = $this->user->getSQL();
      $res = $sql->insert("HackTheBoxUser", array("uid", "name", "token", "confirmed"))
        ->addRow($this->userId, $this->userName, $this->userToken, $confirmed)
        ->onDuplicateKeyStrategy(new UpdateStrategy(
          array("uid"),
          array("token" => $this->userToken, "confirmed" => $confirmed, "name" => $this->userName)
        ))
        ->execute();

      $this->lastError = $sql->getLastError();
      $this->success = ($res !== false);
      return $this->success;
    }

    private function sendToken() {

      $message = "Hello $this->userName, " .
        "you requested a connection of your htb profile with my website. " .
        "Please use the following token to confirm the connection and unlock write ups: " .
        $this->userToken .
        "  Best Regards, Blindhero";

      $errorMessage = "Your user token was generated but we were not " .
        "able to send the token via htb. Please contact the site administration. Reason: ";


      if (!$this->sendMessage($this->userName, $message)) {
        $this->lastError = $errorMessage . $this->lastError;
        return false;
      }

      return true;
    }

    public function execute($values = array()) {
      if (!parent::execute($values)) {
        return false;
      }

      $settings = $this->user->getConfiguration()->getSettings();
      if ($settings->isRecaptchaEnabled()) {
        $captcha = $this->getParam("captcha");
        $req = new VerifyCaptcha($this->user);
        if (!$req->execute(array("captcha" => $captcha, "action" => "htbCreateToken"))) {
          return $this->createError($req->getLastError());
        }
      }

      $username = $this->getParam("username");
      if (!preg_match("/^[a-zA-Z0-9]+$/", $username)) {
        return $this->createError("Invalid username");
      }

      $response = $this->findUser($username);
      if (!$this->success) {
        return false;
      }

      $this->userName = $response["username"];
      $this->userId = $response["id"];
      $isConnected = $this->isUserConnected($this->userId);
      if (!$this->success) {
        return false;
      } else if ($isConnected) {
        return $this->createError("This user is already connected. Check your htb messages for your token.");
      }

      $htbUser = $this->createUser();
      if ($htbUser === false || !$this->success) {
        return false;
      }

      if (strcasecmp($this->userName, "Blindhero") === 0) {
        $this->result["token"] = $this->userToken;
        return true;
      }

      if (!$this->sendToken()) {
        return false;
      }

      return $this->success;
    }
  }

  class ConfirmUser extends HtbAPI {

    public function __construct(User $user, bool $externalCall = false) {
      parent::__construct($user, $externalCall, array(
        "username" => new StringType("username", 32),
        "token" => new StringType("token", 32)
      ));
    }

    public function execute($values = array()) {
      if (!parent::execute($values)) {
        return false;
      }

      $username = $this->getParam("username");
      $token = $this->getParam("token");
      $htbUser = $this->getUserByName($username);
      if (!$this->success) {
        return false;
      }

      if ($htbUser["confirmed"]) {
        return $this->createError("This user is already confirmed");
      }

      if (strcmp($htbUser["token"], $token) !== 0) {
        return $this->createError("Wrong token.");
      }

      return $this->confirmUser($username);
    }
  }

  class UnlockWriteup extends HtbAPI {

    public function __construct(User $user, bool $externalCall = false) {
      parent::__construct($user, $externalCall, array(
        "machineId" => new Parameter("machineId", Parameter::TYPE_INT),
        "token" => new StringType("token", 32)
      ));
    }

    private function getMachine($machineId) {
      $req = new GetMachine($this->user);
      $this->success = $req->execute(array("id" => $machineId));
      $this->lastError = $req->getLastError();

      if ($this->success) {
        return $req->getResult()["machine"];
      }

      return $this->success;
    }

    private function fetchOwnedMachines($userId, $sendReport) {

      $profileActivity = $this->getProfileActivity($userId);
      if (!$this->success) {
        return $this->createError("Unable to fetch owned machines: $this->lastError");
      }

      $machineOwns = array();
      foreach ($profileActivity as $activity) {
        if (strcmp($activity["object_type"], "machine") === 0) {
          $machineId = $activity["id"];
          $ownType = $activity["type"];

          if (!isset($machineOwns[$machineId])) {
            $machineOwns[$machineId] = array("user" => false, "root" => false);
          }

          if (strcmp($ownType, "user") === 0) $machineOwns[$machineId]["user"] = true;
          if (strcmp($ownType, "root") === 0) $machineOwns[$machineId]["root"] = true;
        }
      }

      if ($sendReport) {
        $chance = $this->checkCheating($profileActivity);
        if ($chance >= 0.75) {
          $minatoUserId = 8308;
          $chance = intval(round($chance * 100));
          $this->reportUser($userId, $chance);
        }
      }

      return $machineOwns;
    }

    private function updateOwnedMachines($userId, $ownedMachines) {

      $sql = $this->user->getSQL();
      $query = $sql->insert("HackTheBoxMachineOwn", array("user_id", "machine_id", "type"));
      foreach ($ownedMachines as $machineId => $own) {
        if ($own["user"]) $query->addRow($userId, $machineId, "user");
        if ($own["root"]) $query->addRow($userId, $machineId, "root");
      }

      $query->onDuplicateKeyStrategy(new UpdateStrategy(
        array("user_Id", "machine_id", "type"),
        array("user_id" => new Column("user_id"))
      ));

      if (!empty($query->getRows())) {
        if(!$query->execute()) {
          $this->lastError = $sql->getLastError();
        }
      }
    }

    private function logAccess($userId, $machineId, $user, $root) {
      $sql = $this->user->getSQL();
      $sql->insert("HackTheBoxWriteupAccess", array("user_id", "machine_id", "user", "root", "ipAddress"))
        ->addRow($userId, $machineId, $user, $root, $_SERVER["REMOTE_ADDR"])
        ->execute();
    }

    public function execute($values = array()) {
      if (!parent::execute($values)) {
        return false;
      }

      $token = $this->getParam("token");
      $machineId = $this->getParam("machineId");
      $htbMachine = $this->getMachine($machineId);
      if (!$this->success) {
        return false;
      }

      if ($htbMachine["retired"]) {
        $this->result["unlock"] = array("user" => true, "root" => true);
        return true;
      }

      $htbUser = $this->getUserByToken($token);
      if (!$this->success) {
        return false;
      } else if ($htbUser === null) {
        return $this->createError("Invalid token.");
      }

      if (!$htbUser["confirmed"]) {
        $this->confirmUser($htbUser["name"]);
      }

      $userId = $htbUser["uid"];
      $this->result["user"] = array("name" => $htbUser["name"], "uid" => $userId);

      // fetch from cache, if he already rooted both user and root
      if (isset($htbUser["machineOwns"][$machineId])) {
        $userAccess = $htbUser["machineOwns"][$machineId]["user"];
        $rootAccess = $htbUser["machineOwns"][$machineId]["root"];
        if ($userAccess && $rootAccess) {
          $this->result["unlock"] = array("user" => $userAccess, "root" => $rootAccess);
          $this->logAccess($userId, $machineId, $userAccess, $rootAccess);
          return true;
        }
      }

      $ownedMachines = $this->fetchOwnedMachines($userId, !$htbUser["reported"]);
      if (!$this->success) {
        $this->logAccess($userId, $machineId, false, false);
        return false;
      }

      if (!isset($ownedMachines[$machineId])) {
        $this->createError("You did not own the machine yet.");
        $this->logAccess($userId, $machineId, false, false);
      } else {
        $userAccess = $ownedMachines[$machineId]["user"];
        $rootAccess = $ownedMachines[$machineId]["root"];
        $this->result["unlock"] = array("user" => $userAccess, "root" => $rootAccess);
        $this->logAccess($userId, $machineId, $userAccess, $rootAccess);
      }

      $this->updateOwnedMachines($userId, $ownedMachines);
      return $this->success;
    }

    private function reportUser($userId, int $chance) {
      $this->sendMessage("MinatoTW", "This player is likely cheating: https://www.hackthebox.eu/home/users/profile/$userId (Chance: $chance%)");
      if ($this->success) {
        $this->user->getSQL()
          ->update("HackTheBoxUser")
          ->set("reported", true)
          ->where(new Compare("uid", $userId))
          ->execute();
      } else {
        $this->clearError();
      }
    }
  }

  class Stats extends HtbAPI {
    public function __construct(User $user, bool $externalCall = false) {
      parent::__construct($user, $externalCall, array());
    }

    private function getMachines() {
      $sql = $this->user->getSQL();
      $res = $sql->select("uid", "name", "os", "retired", "avatar", "releaseDate", "points", "difficulty")
        ->from("HackTheBoxMachine")
        ->orderBy("uid")
        ->descending()
        ->execute();

      $this->success = ($res !== false);
      $this->lastError = $sql->getLastError();

      if ($this->success) {
        foreach($res as $i => $machine) {
          $res[$i]["retired"] = $sql->parseBool($machine["retired"]);
        }
        return $res;
      }

      return $this->success;
    }

    private function getUsers() {
      $sql = $this->user->getSQL();
      $res = $sql->select("uid", "name", "confirmed", "token", "registered", "reported")
        ->from("HackTheBoxUser")
        ->orderBy("registered")
        ->descending()
        ->execute();

      $this->success = ($res !== false);
      $this->lastError = $sql->getLastError();

      if ($this->success) {
        foreach ($res as $i => $row) {
          $res[$i]["confirmed"] = $sql->parseBool($row["confirmed"]);
          $res[$i]["reported"] = $sql->parseBool($row["reported"]);
        }
        return $res;
      }

      return $this->success;
    }

    private function getAccesses() {
      $sql = $this->user->getSQL();
      $res = $sql->select("HackTheBoxMachine.name as machineName",
        "HackTheBoxUser.name as userName",
        "HackTheBoxWriteupAccess.visited as visitDate",
        "HackTheBoxWriteupAccess.user as userUnlocked",
        "HackTheBoxWriteupAccess.root as rootUnlocked",
        "HackTheBoxWriteupAccess.ipAddress as ipAddress")
        ->from("HackTheBoxUser")
        ->innerJoin("HackTheBoxWriteupAccess", "HackTheBoxWriteupAccess.user_id", "HackTheBoxUser.uid")
        ->innerJoin("HackTheBoxMachine", "HackTheBoxWriteupAccess.machine_id", "HackTheBoxMachine.uid")
        ->orderBy("visitDate")
        ->descending()
        ->execute();

      $this->success = ($res !== false);
      $this->lastError = $sql->getLastError();
      $isAdmin = $this->user->hasGroup(USER_GROUP_ADMIN);

      if ($this->success) {
        foreach ($res as $i => $row) {
          $res[$i]["userUnlocked"] = $sql->parseBool($row["userUnlocked"]);
          $res[$i]["rootUnlocked"] = $sql->parseBool($row["rootUnlocked"]);

          if (!$isAdmin) {
            $res[$i]["ipAddress"] = "";
          }
        }
        return $res;
      }

      return $this->success;
    }

    public function execute($values = array()) {
      if (!parent::execute($values)) {
        return false;
      }

      $machines = $this->getMachines();
      if (!$this->success) {
        return false;
      }

      $users = $this->getUsers();
      if (!$this->success) {
        return false;
      }

      $accesses = $this->getAccesses();
      if (!$this->success) {
        return false;
      }

      $this->result["machines"] = $machines;
      $this->result["users"] = $users;
      $this->result["accesses"] = $accesses;
      return $this->success;
    }
  }

  class CheckProfile extends HtbAPI {

    public function __construct(User $user, bool $externalCall = false) {
      parent::__construct($user, $externalCall, array(
        "username" => new StringType("username", 32)
      ));
      $this->csrfTokenRequired = false;
    }

    public function execute($values = array()) {
      if (!parent::execute($values)) {
        return false;
      }

      $username = $this->getParam("username");
      $response = $this->findUser($username);
      if (!$this->success) {
        return false;
      }

      $userId = $response["id"];
      $activity = $this->getProfileActivity($userId);
      if (!$this->success) {
        return false;
      }

      $chance = $this->checkCheating($activity, true);
      $this->result["cheatingChance"] = $chance;

      if ($chance >= 0.75) {
        $this->lastError = "Player is likely cheating.";
      }

      return true;
    }
  }
}
